(ns puppetlabs.rbac.services.http.api
  (:require [clojure.java.io :as io]
            [clojure.string  :as str]

            [cheshire.core            :as json]
            [clj-jwt.core             :as jwt]
            [compojure.core           :refer [routes GET POST ANY] :as c]

            [compojure.handler        :as handler]
            [compojure.route          :as route]
            [liberator.core           :refer [defresource by-method]]
            [liberator.dev            :refer [wrap-trace]]
            [liberator.representation :refer [ring-response]]
            [ring.util.servlet        :refer [servlet]]
            [schema.core              :as sc]
            [slingshot.slingshot      :refer [throw+]]

            [puppetlabs.kitchensink.core                  :refer [parse-int]]
            [puppetlabs.rbac.http.middleware              :refer [wrap-errors]]
            [puppetlabs.rbac.ldap                         :as ldap]
            [puppetlabs.rbac.schema                       :as rs]
            [puppetlabs.rbac.services.rbac                :as rbac]
            [puppetlabs.rbac.services.http.api-validation :as val]
            [puppetlabs.rbac.shiro-utils                  :as shiro]
            [puppetlabs.rbac.storage                      :as st]
            [puppetlabs.rbac.storage.permissioned         :as pm]
            [puppetlabs.trapperkeeper.core                :refer [defservice]]
            [puppetlabs.trapperkeeper.services            :refer [service-context get-service]]))

(defn json-content? [ctx]
  (if-let [content-type (get-in ctx [:request :headers "content-type"])]
    (re-matches #"application/json.*" content-type)
    false))

(defn sec-mgr [ctx]
  (get-in ctx [:request :sec-mgr]))

(defn shiro-subject [ctx]
  (get-in ctx [:request :shiro-subject]))

(defn rbac-user
  "Use the currently authenticated user from `context` to retrieve
   the rbac user"
  [db context]
  (->> (shiro-subject context)
       .getPrincipal
       (st/get-user db :login)))

(def standard-json-resource
  {:available-media-types ["application/json"]
   :handle-malformed val/handle-malformed })

(defresource subject-get-permissions [db]
  standard-json-resource
  :allowed-methods        [:post]
  :malformed?             #(val/extract-and-validate rs/PermissionsQuery :body %)
  :known-content-type?    json-content?
  :exists?                true
  :new?                   false
  :respond-with-entity?   true
  :handle-ok              (fn [ctx] (st/get-permissions db (parse-int (:token (:body ctx)))
                                                          (:permissions (:body ctx)))))

(defresource subject-check-permission [db sid perm-str]
  standard-json-resource
  :allowed-methods       [:get]
  :exists?               (fn [ctx] (when-let [subject (st/get-subject db :id sid)]
                                     {:subject (shiro/login->subject (sec-mgr ctx) (:login subject))}))
  :handle-ok             (fn [ctx] {:granted (shiro/is-permitted? (:subject ctx) perm-str)}))

(defresource permitted [db]
  standard-json-resource
  :allowed-methods [:post]
  :malformed? (fn [ctx] (val/extract-and-validate rs/PermittedQuery :body ctx))
  :exists? (fn [ctx] (when-let [subject (->> (parse-int (:token (:body ctx)))
                                             (st/get-subject db :id))]
                       {:subject (shiro/login->subject (sec-mgr ctx) (:login subject))}))
  :new? false
  :respond-with-entity? true
  :handle-ok (fn [ctx]
               (let [perm-strs (map rs/permission->string (:permissions (:body ctx)))]
                 (shiro/are-permitted? (:subject ctx) perm-strs))))

(defresource users [db api-root id-filter]
  standard-json-resource
  :allowed-methods       [:get :post]
  :malformed?            (by-method {:get  (fn [_] (val/extract-id-filter id-filter))
                                     :post #(val/extract-and-validate rs/PostUser :user %)})
  :known-content-type?   (by-method {:get  true
                                     :post json-content?})
  :exists?               (by-method {:get  true
                                     :post #(when-let [record (st/get-subject db "login" (get-in % [:user :login]))]
                                              {:user-exists true})})
  :handle-ok             (fn [ctx] (st/get-users db (:id-filter ctx)))
  :post!                 #(when-not (contains? % :user-exists)
                            (st/create-local-user! db (:user %)))
  :post-redirect?        true
  :handle-see-other      #(if (contains? % :user-exists)
                            (ring-response {:status 409})
                            (ring-response {:status 303
                                            :headers {"Location" (format "%s/users/%s"
                                                                         api-root
                                                                         (:id (st/get-subject db "login" (get-in % [:user :login]))))}})))

(defresource user [db id]
  standard-json-resource
  :allowed-methods        [:get :put]
  :malformed?             (by-method {:get false
                                      :put #(val/extract-and-validate rs/User :modified-user % id)})
  :known-content-type?    (by-method {:get true
                                      :put json-content?})
  :exists?                (fn [_] (when-let [record (st/get-user db "id" id)] {:user record}))
  :conflict?              #(let [modified-user (:modified-user %)]
                             (when-not (:is_remote modified-user)
                               (if-let [possible-dupe (st/get-user db "login" (:login modified-user))]
                                 (not= (:id possible-dupe) id)
                                 false)))
  :handle-ok              (by-method {:get :user
                                      :put (fn [_] (st/get-user db "id" id))})
  :can-put-to-missing?    false
  :handle-not-implemented (ring-response {:status 404}) ; this returns a 404 for a request to PUT to a resource that doesn't exist
  :put!                   #(st/update-user! db id (:modified-user %))
  :new?                   false
  :respond-with-entity?   true)

(defresource current-user [db]
  standard-json-resource
  :allowed-methods        [:get]
  :exists?                (fn [ctx]
                            (when-let [user (rbac-user db ctx)]
                              {:user user}))
  :handle-ok              :user)

(defresource user-password [rbac-svc]
  standard-json-resource
  :allowed-methods        [:put]
  :malformed?            #(val/extract-and-validate rs/ChangePassword :user-password %)
  :allowed?               (fn [ctx]
                            (let [{:keys [login is_remote]} (rbac-user (:db (service-context rbac-svc)) ctx)]
                              (cond
                               is_remote
                               [false {:forbidden-msg "Password change not allowed on remote users"}]

                               (not (rbac/password-matches? rbac-svc login (get-in ctx [:user-password :current_password])))
                               [false {:forbidden-msg "Incorrect current password provided"}]

                               :else
                               [true {:user-login login}])))

  :handle-forbidden       :forbidden-msg
  :known-content-type?    json-content?
  :put!                   (fn [ctx]
                            (rbac/update-local-user-password! rbac-svc
                                                              (:user-login ctx)
                                                              (get-in ctx [:user-password :password])
                                                              false))
  :new?                   false
  :respond-with-entity?   true)

(defresource password-reset [rbac-svc sid]
  standard-json-resource
  :allowed-methods       [:post]
  :allowed?              (fn [context]
                           (let [requesting-user (rbac-user (:db (service-context rbac-svc)) context)
                                 user-sid (parse-int sid)
                                 user-being-changed (st/get-user (:db (service-context rbac-svc)) :id user-sid)]
                             (cond
                              ;; If user is not found, fall to exists?
                              ;; and handle it there
                              (and user-being-changed (:is_remote user-being-changed))
                              [false {:forbidden-msg "Password reset not allowed on remote users"}]

                              (not (shiro/is-permitted? (shiro-subject context) "users:reset_password"))
                              [false {:forbidden-msg "Insufficient permissions to reset a user's password"}]

                              :else
                              {:user-being-changed user-being-changed
                               :user-sid user-sid
                               :requesting-user requesting-user})))
  :handle-forbidden      :forbidden-msg
  :exists?               (fn [{:keys [user-being-changed]}]
                           (if user-being-changed
                             true
                             [false (format "User %s not found" sid)]))
  :post!                 (fn [{:keys [requesting-user user-sid]}]
                           {:token (rbac/reset-password-token! rbac-svc
                                                               (:login requesting-user)
                                                               user-sid)})
  :new?                  true
  :handle-created        :token
  :post-redirect?        false)

(defresource change-password-with-token [rbac-svc]
  :allowed-methods        [:post]
  :malformed?             #(val/extract-and-validate rs/ChangePasswordWithToken :change-password-data %)

  :handle-malformed       :error
  :allowed?               (fn [ctx]
                            (let [token (jwt/str->jwt (get-in ctx [:change-password-data :token]))]
                              (if-let [validation-msg (rbac/validate-reset-token rbac-svc token)]
                                [false {:forbidden-msg validation-msg}]
                                (if-let [subject (rbac/get-subject rbac-svc :reset_password_uuid (java.util.UUID/fromString (get-in token [:claims :sub])))]
                                  [true {:user-from-token subject}]
                                  [false {:forbidden-msg "Password reset failed, token has already been used"}]))))

  :handle-forbidden       :forbidden-msg
  :known-content-type?    json-content?
  :available-media-types  ["application/json"]
  :post!                  (fn [ctx]
                            (rbac/update-local-user-password! rbac-svc
                                                              (get-in ctx [:user-from-token :login])
                                                              (get-in ctx [:change-password-data :password])
                                                              true))
  :new?                   false
  :respond-with-entity?   true)

(defresource groups [db api-root id-filter]
  standard-json-resource
  :allowed-methods       [:get :post]
  :malformed?            (by-method {:get  (fn [_] (val/extract-id-filter id-filter))
                                     :post #(val/extract-and-validate rs/PostGroup :group %)})
  :known-content-type?   (by-method {:get  true
                                     :post json-content?})
  :exists?               (by-method {:get  true
                                     :post #(when-let [record (st/get-group db "login" (get-in % [:group :login]))]
                                              {:group-exists true})})
  :handle-ok             (fn [ctx] (pm/get-groups (shiro-subject ctx) db (:id-filter ctx)))
  :post!                 #(when-not (contains? % :group-exists)
                            (pm/import-group! (shiro-subject %) db (:group %)))
  :post-redirect?        true
  :handle-see-other      #(if (contains? % :group-exists)
                            (ring-response {:status 409})
                            (ring-response {:status 303
                                            :headers {"Location" (format "%s/groups/%s"
                                                                         api-root
                                                                         (:id (st/get-group db "login" (get-in % [:group :login]))))}})))

(defresource group [db id]
  standard-json-resource
  :allowed-methods        [:get :put]
  :malformed?             (by-method {:get false
                                      :put #(val/extract-and-validate rs/RemoteGroup :modified-group % id)})
  :known-content-type?    (by-method {:get true
                                      :put json-content?})
  :exists?                (fn [ctx] (when-let [record (pm/get-group (shiro-subject ctx) db "id" id)] {:group record}))
  :handle-ok              (by-method {:get :group
                                      :put (fn [ctx] (pm/get-group (shiro-subject ctx) db "id" id))})
  :can-put-to-missing?    false
  :handle-not-implemented (ring-response {:status 404}) ; this returns a 404 for a request to PUT to a resource that doesn't exist
  :put!                   #(pm/update-group! (shiro-subject %) db id (:modified-group %))
  :new?                   false
  :respond-with-entity?   true)

(defresource roles [db api-root id-filter]
  standard-json-resource
  :allowed-methods       [:get :post]
  :malformed?            (by-method {:get  (fn [_] (val/extract-id-filter id-filter))
                                     :post #(val/extract-and-validate rs/Role :role %)})
  :known-content-type?   (by-method {:get  true
                                     :post json-content?})
  :exists?               (by-method {:get  true
                                     :post #(when (st/get-role db "display_name" (get-in % [:role :display_name]))
                                              {:role-exists true})})
  :handle-ok             (fn [ctx] (st/get-roles db (:id-filter ctx)))
  :post!                 #(when-not (contains? % :role-exists)
                            (st/create-role! db (:role %)))
  :post-redirect?        true
  :handle-see-other      #(if (contains? % :role-exists)
                            (ring-response {:status 409})
                            (ring-response {:status 303
                                            :headers {"Location" (format "%s/roles/%s"
                                                                         api-root
                                                                         (:id (st/get-role db "display_name" (get-in % [:role :display_name]))))}})))

(defresource role [db id]
  standard-json-resource
  :allowed-methods        [:get :put :delete]
  :malformed?             (by-method {:get  false
                                      :delete false
                                      :put #(val/extract-and-validate rs/Role :modified-role % id)})
  :known-content-type?    (by-method {:get true
                                      :put json-content?
                                      :delete json-content?})
  :exists?                (fn [_] (when-let [record (st/get-role db "id" id)] {:role record}))
  :conflict?              #(let [modified-role (:modified-role %)]
                             (if-let [possible-dupe (st/get-role db "display_name" (:display_name modified-role))]
                                 (not= (:id possible-dupe) id)
                                 false))
  :handle-ok              (by-method {:get :role
                                      :put (fn [_] (st/get-role db "id" id))
                                      :delete (fn [ctx] (dissoc (:role ctx) :id))})
  :delete!                (fn [_] (st/delete-role! db id))
  :can-put-to-missing?    false
  :handle-not-implemented (ring-response {:status 404}) ; this returns a 404 for a request to PUT to a resource that doesn't exist
  :put!                   #(st/update-role! db id (:modified-role %))
  :new?                   false
  :respond-with-entity?   true)

(defresource types [db]
  standard-json-resource
  :allowed-methods [:get]
  :handle-ok (fn [_] (st/get-types db)))

(defresource directory-settings [rbac-svc]
  standard-json-resource
  :allowed-methods       [:get :put]
  :malformed?            (by-method {:get false
                                     :put #(val/directory-settings-validate :ds-settings %)})
  :known-content-type?   (by-method {:get true
                                     :put json-content?})
  :handle-ok             (fn [ctx] (rbac/get-directory-settings rbac-svc (shiro-subject ctx)))
  :new?                  false
  :respond-with-entity?  true
  :put!                  (fn [ctx] (rbac/update-directory-settings! rbac-svc (shiro-subject ctx) (:ds-settings ctx))))

(defresource test-directory-settings [rbac-svc]
  standard-json-resource
  :allowed-methods       [:get :put]
  :malformed?            (by-method {:get false
                                     :put #(val/directory-settings-validate :ds-settings %)})
  :exists?               (by-method {:get (fn [ctx] {:ds-settings (rbac/get-directory-settings rbac-svc (shiro-subject ctx))})
                                     :put :ds-settings})
  :handle-ok             (fn [ctx]
                           (let [settings (:ds-settings ctx)
                                 result   (pm/test-connection (shiro-subject ctx) settings)]
                             (dissoc result :connection)))
  :respond-with-entity?  true
  :new?                  false)


(defn unauthenticated-routes [rbac-svc]
  (-> (routes
       (ANY "/auth/reset" []
            (change-password-with-token rbac-svc)))
      handler/api))

(defn ds-api-routes [rbac-svc]
  (-> (routes
        (ANY "/ds" []
             (directory-settings rbac-svc))

        (ANY "/ds/test" []
             (test-directory-settings rbac-svc)))
      handler/api))

(defn pw-api-routes [rbac-svc]
  (-> (routes
        (ANY "/users/current/password" []
             (user-password rbac-svc))

        (ANY "/users/:sid/password/reset" [sid]
             (password-reset rbac-svc sid)))

      handler/api))

(defn api-routes [db api-root]
  (-> (routes
        (ANY "/permissions" []
             (subject-get-permissions db))

        (ANY "/permitted" []
             (permitted db))

        (ANY "/subjects/:sid/:perm-str" [sid perm-str]
             (subject-check-permission db (parse-int sid) perm-str))

        (ANY ["/users/:sid", :sid #"[0-9]+"] [sid]
             (user db (parse-int sid)))

        (ANY "/users" [id] ;; id is an id filter (potentially a CSL)
             (users db api-root id))

        (ANY "/users/current" []
             (current-user db))

        (ANY "/groups/:sid" [sid]
             (group db (parse-int sid)))

        (ANY "/groups" [id]
             (groups db api-root id))

        (ANY "/roles/:rid" [rid]
             (role db (parse-int rid)))

        (ANY "/roles" [id]
             (roles db api-root id))

        (ANY "/types" []
             (types db)))

      handler/api))

(defn web-app
  [api-path login-path rbac-svc wrap-authenticated wrap-unauthenticated]
  (let [unauth-routes (wrap-unauthenticated (unauthenticated-routes rbac-svc))
        auth-routes (wrap-authenticated
                      (routes
                        (api-routes (:db (service-context rbac-svc))
                                    (str api-path "/v1"))
                        (ds-api-routes rbac-svc)
                        (pw-api-routes rbac-svc)))
        v1-api (c/context "/v1" []
                          (routes unauth-routes
                                  auth-routes
                                  (route/not-found "Not Found")))]
    (c/context api-path []
               (-> v1-api
                   (wrap-trace :header :ui) ;; TODO remove before we ship))
                   (wrap-errors login-path)))))

(defservice rbac-http-api-service
  [[:ConfigService get-in-config]
   [:WebroutingService add-servlet-handler get-route]
   [:RbacAuthnMiddleware wrap-authenticated wrap-unauthenticated]
   RbacService]
  (init [this tk-ctx]
        (let [rbac-svc (get-service this :RbacService)
              api-path (get-route this)
              login-path (get-in-config [:global :login-path])
              app (web-app api-path login-path rbac-svc wrap-authenticated wrap-unauthenticated)]

          ;; Shiro authn requires use of servlet
          (add-servlet-handler this (servlet app))

          tk-ctx)))
