(ns puppetlabs.redrover.file
  (:require [cheshire.core :as json]
            [clj-yaml.core :as yaml]
            [clojure.string :as str]
            [slingshot.slingshot :refer [try+ throw+]]))

(defn- filenotfound-reason
  [e]
  (let [maybe-err-reason (->> e
                           str
                           (re-find #"\((.+?)\)")
                           last)]
    (if-let [reason (not-empty maybe-err-reason)]
      (str/lower-case reason)
      ;; This shouldn't ever happen, but in case it does
      "unknown")))

(defn spit-json
  [path data]
  (try (spit path (json/encode data {:pretty true}))
    (catch java.io.FileNotFoundException e
      (throw+ {:kind ::path-not-writable
               :msg (str "Cannot write to '" path "' for the following reason: "
                         (filenotfound-reason e))}))))

(defn read-file
  "Equivalent to slurp, but will catch java.io exceptions and throw
  ::path-not-readable"
  [path]
  (try (slurp path)
    (catch java.io.FileNotFoundException e
      (throw+ {:kind ::path-not-readable
               :msg (str "Cannot read from '" path "' for the following reason: "
                         (filenotfound-reason e))}))))

(defn read-yaml-file
  [path]
  (try (yaml/parse-string (read-file path))
    (catch org.yaml.snakeyaml.error.YAMLException e
      (throw+ {:kind ::malformed-yaml
               :msg (str "The file '" path "' does not appear to contain valid YAML.")}))))

(defn read-json-file
  ([path]
   (let [file-contents (read-file path)]
     (try (json/decode file-contents true)
       (catch com.fasterxml.jackson.core.JsonProcessingException e
         (throw+ {:kind ::malformed-json
                  :msg (str "The file '" path "' does not appear to contain valid JSON.")}))))))
